;---------------------------------------
; Commodore 64 emulator
;
; Translated and corrected sourcecodes
; originally by unknown author
;
; Version 0.2
;---------------------------------------
; Emulates a 6502 (without secret opcodes and extra 6510 opcodes, just standard C64 opcodes)
; Does not emulate SID or Joystickports, userport, cartridge, disk, datasette, CIA(input/output 2 6526 adapters w16bit parallel IO 8bit serial IO, clock with alarm)
; Does an simple VIC II emulation (40x25 textmode 1024 area without colourmap, and vbl interrupt) nothing else
; Notes: 100 - 1FF is stack page
; Improved C64 palette: black, white, red, cyan, purple, green, blue, yellow, lt brown, brown, pink, dark gray, gray, lt green, lt blue, lt gray
; $000000, $FFFFFF, $68372B, $70A4B2, $6F3D86, $588D43, $352879, $B8C76F, $6F4F25, $433900, $9A6759, $444444, $6C6C6C, $9AD284, $6C5EB5, $959595

Graphics 1024,600

; Global variables
Global a,x,y,s,p,op   ; Bytes
Global pc         ; Words
Dim ram(65536)      ; Arrays are global
Dim basic_rom(65536); $A000-$BFFF
Dim char_rom(65536) ; $D000-$DFFF
Dim kernal_rom(65536);$E000-$FFFF

; Load roms
f=ReadFile("BASIC.ROM") : For temp=0 To 8191 : byte=ReadByte(f) : basic_rom($A000+temp)=byte : Next : CloseFile f
f=ReadFile("CHAR.ROM") : For temp=0 To 4095 : byte=ReadByte(f) : char_rom($D000+temp)=byte : Next : CloseFile f
f=ReadFile("KERNAL.ROM") : For temp=0 To 8191 : byte=ReadByte(f) : kernal_rom($E000+temp)=byte : Next : CloseFile f

; Constants
Const c=1, z=2, i=4, d=8, b=16, v=64, n=128 ; statusreg p: NV.BDIZC

; Processor

reset()

; Main loop  OKAY
Repeat
cycle=cycle+1 : op=peek(pc) : pc=pc+1

Select op
           Case $69 : adc(imm()) ; OKAY
           Case $65 : adc(zp()) ; OKAY
           Case $75 : adc(zpx()) ; OKAY
           Case $6D : adc(abs_()) ; OKAY
           Case $7D : adc(abs_x()) ; OKAY
           Case $79 : adc(abs_y()) ; OKAY
           Case $61 : adc(indx()) ; OKAY
           Case $71 : adc(indy()) ; OKAY
           Case $29 : and_(imm()) ; OKAY
           Case $25 : and_(zp()) ; OKAY
           Case $35 : and_(zpx()) ; OKAY
           Case $2D : and_(abs_()) ; OKAY
           Case $3D : and_(abs_x()) ; OKAY
           Case $39 : and_(abs_y()) ; OKAY
           Case $21 : and_(indx()) ; OKAY
           Case $31 : and_(indy()) ; OKAY
           Case $0A : asl_A() ; OKAY
           Case $06 : asl(zp()) ; OKAY
           Case $16 : asl(zpx()) ; OKAY
           Case $0E : asl(abs_()) ; OKAY
           Case $1E : asl(abs_x()) ; OKAY
           Case $90 : bfc(c) ; bcc ; OKAY
           Case $B0 : bfs(c) ; bcs ; OKAY
           Case $F0 : bfs(z) ; beq ; OKAY
           Case $24 : bit(zp()) ; OKAY
           Case $2C : bit(abs_()) ; OKAY
           Case $30 : bfs(n) ; bmi ; OKAY
           Case $D0 : bfc(z) ; bne ; OKAY
           Case $10 : bfc(n) ; bpl ; OKAY
           Case $00 : brk() ; OKAY
           Case $50 : bfc(v) ; bvc ; OKAY
           Case $70 : bfs(v) ; bvs ; OKAY
           Case $18 : setflag(c,False) ; clc ; OKAY
           Case $D8 : setflag(d,False) ; cld ; OKAY
           Case $58 : setflag(i,False) ; cli ; OKAY
           Case $B8 : setflag(v,False) ; clv ; OKAY
           Case $C9 : cmp(imm()) ; OKAY
           Case $C5 : cmp(zp()) ; OKAY
           Case $D5 : cmp(zpx()) ; OKAY
           Case $CD : cmp(abs_()) ; OKAY
           Case $DD : cmp(abs_x()) ; OKAY
           Case $D9 : cmp(abs_y()) ; OKAY
           Case $C1 : cmp(indx()) ; OKAY
           Case $D1 : cmp(indy()) ; OKAY
           Case $E0 : cpx(imm()) ; OKAY
           Case $E4 : cpx(zp()) ; OKAY
           Case $EC : cpx(abs_()) ; OKAY
           Case $C0 : cpy(imm()) ; OKAY
           Case $C4 : cpy(zp()) ; OKAY
           Case $CC : cpy(abs_()) ; OKAY
           Case $C6 : dec_(zp()) ; OKAY
           Case $D6 : dec_(zpx()) ; OKAY
           Case $CE : dec_(abs_()) ; OKAY
           Case $DE : dec_(abs_x()) ; OKAY
           Case $CA : dex() ; OKAY
           Case $88 : dey() ; OKAY
           Case $49 : eor(imm()) ; OKAY
           Case $45 : eor(zp()) ; OKAY
           Case $55 : eor(zpx()) ; OKAY
           Case $4D : eor(abs_()) ; OKAY
           Case $5D : eor(abs_x()) ; OKAY
           Case $59 : eor(abs_y()) ; OKAY
           Case $41 : eor(indx()) ; OKAY
           Case $51 : eor(indy()) ; OKAY
           Case $E6 : inc_(zp()) ; OKAY
           Case $F6 : inc_(zpx()) ; OKAY
           Case $EE : inc_(abs_()) ; OKAY
           Case $FE : inc_(abs_x()) ; OKAY
           Case $E8 : inx() ; OKAY
           Case $C8 : iny() ; OKAY
           Case $4C : jmp(abs_()) ; OKAY
           Case $6C : jmp(ind()) ; OKAY
           Case $20 : jsr(abs_()) ; OKAY
           Case $A9 : lda(imm()) ; OKAY
           Case $A5 : lda(zp()) ; OKAY
           Case $B5 : lda(zpx()) ; OKAY
           Case $AD : lda(abs_()) ; OKAY
           Case $BD : lda(abs_x()) ; OKAY
           Case $B9 : lda(abs_y()) ; OKAY
           Case $A1 : lda(indx()) ; OKAY
           Case $B1 : lda(indy()) ; OKAY
           Case $A2 : ldx(imm()) ; OKAY
           Case $A6 : ldx(zp()) ; OKAY
           Case $B6 : ldx(zpy()) ; OKAY
           Case $AE : ldx(abs_()) ; OKAY
           Case $BE : ldx(abs_y()) ; OKAY
           Case $A0 : ldy(imm()) ; OKAY
           Case $A4 : ldy(zp()) ; OKAY
           Case $B4 : ldy(zpx()) ; OKAY
           Case $AC : ldy(abs_()) ; OKAY
           Case $BC : ldy(abs_x()) ; OKAY
           Case $4A : lsr_A() ; OKAY
           Case $46 : lsr(zp()) ; OKAY
           Case $56 : lsr(zpx()) ; OKAY
           Case $4E : lsr(abs_()) ; OKAY
           Case $5E : lsr(abs_x()) ; OKAY
           Case $EA : ; nop ; OKAY
           Case $09 : ora(imm()) ; OKAY
           Case $05 : ora(zp()) ; OKAY
           Case $15 : ora(zpx()) ; OKAY
           Case $0D : ora(abs_()) ; OKAY
           Case $1D : ora(abs_x()) ; OKAY
           Case $19 : ora(abs_y()) ; OKAY
           Case $01 : ora(indx()) ; OKAY
           Case $11 : ora(indy()) ; OKAY
           Case $48 : push(a)  ; pha ; OKAY
           Case $08 : push(p)  ; php ; OKAY
           Case $68 : pla() ; OKAY
           Case $28 : p=pull() ; plp ; OKAY
           Case $2A : rol_A() ; OKAY
           Case $26 : rol(zp()) ; OKAY
           Case $36 : rol(zpx()) ; OKAY
           Case $2E : rol(abs_()) ; OKAY
           Case $3E : rol(abs_x()) ; OKAY
           Case $6A : ror_A() ; OKAY
           Case $66 : ror(zp()) ; OKAY
           Case $76 : ror(zpx()) ; OKAY
           Case $6E : ror(abs_()) ; OKAY
           Case $7E : ror(abs_x()) ; OKAY
           Case $40 : rti() ; OKAY
           Case $60 : rts() ; OKAY
           Case $E9 : sbc(imm()) ; OKAY
           Case $E5 : sbc(zp()) ; OKAY
           Case $F5 : sbc(zpx()) ; OKAY
           Case $ED : sbc(abs_()) ; OKAY
           Case $FD : sbc(abs_x()) ; OKAY
           Case $F9 : sbc(abs_y()) ; OKAY
           Case $E1 : sbc(indx()) ; OKAY ; OKAY
           Case $F1 : sbc(indy()) ; OKAY
           Case $38 : setflag(c,True)     ; sec ; OKAY
           Case $F8 : setflag(d,True)     ; sed ; OKAY
           Case $78 : setflag(i,True)     ; sei ; OKAY
           Case $85 : sta(zp()) ; OKAY
           Case $95 : sta(zpx()) ; OKAY
           Case $8D : sta(abs_()) ; OKAY
           Case $9D : sta(abs_x()) ; OKAY
           Case $99 : sta(abs_y()) ; OKAY
           Case $81 : sta(indx()) ; OKAY
           Case $91 : sta(indy()) ; OKAY
           Case $86 : stx(zp()) ; OKAY
           Case $96 : stx(zpy()) ; OKAY
           Case $8E : stx(abs_()) ; OKAY
           Case $84 : sty(zp()) ; OKAY
           Case $94 : sty(zpx()) ; OKAY
           Case $8C : sty(abs_()) ; OKAY
           Case $AA : tax() ; OKAY
           Case $A8 : tay() ; OKAY
           Case $BA : tsx() ; OKAY
           Case $8A : txa() ; OKAY
           Case $9A : txs() ; OKAY
           Case $98 : tya() ; OKAY
           Default Print "unknown instruction":WaitKey:End
End Select

If cycle>1000000/50 Then cycle=cycle-1000000/50:VWait:irq() ; 1000000 instructions / vertical blanking
Until MouseDown(2) ; quit when pressing right mousebutton

Function peek2(address) : peek2 = peek(address) + peek(address+1)*256 : Return peek2 : End Function ; OKAY

Function poke(address,value) ; pokes a byte
ram(address)=value : If address=>$400 And address=<$7E7 Then VicText(address-$400,value)
End Function

Function peek(address) ; returns a byte ; OKAY
peek=ram(address)
If address=>$A000 And address=<$BFFF Then peek=basic_rom(address)
If address=>$D000 And address=<$DFFF Then peek=io(address)
If address=>$E000 And address=<$FFFF Then peek=kernal_rom(address)
Return peek
End Function

Function io(address) ; OKAY
If address = $D012 Then io=$00 Else io=$FF ; fake rasterinterrupt (rasterline 0 or 256)
Return io
End Function

Function signedbyte(byte) ; Checked, reads byte as 2nd complement byte, outputs -128 to 127
temp=byte:If temp=>128 Then temp=temp-256
Return temp
End Function

; Addressing modes: functions which return address for instructions to work upon ; 99%
Function imm() : imm=pc : pc=pc+1 : pc=pc And 65535 : Return imm : End Function ; returns normal word
Function zp() : zp=peek(pc) : pc=pc+1 : pc = pc And 65535 : Return zp : End Function ; returns normal byte
Function zpx() : zpx=peek(pc)+x : zpx = zpx and 255 : pc=pc+1 : pc=pc And 65535 : Return zpx : End Function
Function zpy() : zpy=peek(pc)+y : zpy = zpy and 255 : pc=pc+1 : pc=pc And 65535 : Return zpy : End Function
Function abs_() : abs_=peek2(pc) : pc=pc+2 : abs_=Abs_ And 65535 : pc = pc And 65535 : Return abs_ : End Function
Function abs_x() : absx=peek2(pc)+x : absx = absx and 65535 : pc=pc+2 : pc = pc and 65535 : Return absx : End Function
Function abs_y() : absy=peek2(pc)+y : absy = absy and 65535 : pc=pc+2 : pc = pc and 65535 : Return absy : End Function
Function ind() : ind=peek2(peek2(pc)) : pc=pc+2 : pc = pc and 65535 : Return ind : End Function
Function indx() : indx=peek2(lo(peek(pc)+x)) : indx = indx And 65535 : pc=pc+1 : pc = pc And 65535 : Return indx : End Function
Function indy() : indy=peek2(peek(pc))+y : indy = indy and 65535 : pc=pc+1 : pc = pc and 65535 : Return indy : End Function

Function setflag(flag,status) ; OKAY
If status Then p=(p Or flag) Else p=(p And (flag Xor 255))
End Function

Function flagset(flag) : flagset=(p And flag) : Return flagset : End Function ; OKAY
Function lo(value) : Return value And 255 : End Function ; OKAY
Function hi(value) : value=value Shr 8 : Return value And 255 : End Function
Function push(byte) : poke (256+s,byte) : s=s-1 : s=Abs(s+256) And 255 : End Function ; OKAY
Function pull() : s=s+1 : s=Abs(s+256) And 255 : pull=peek(256+s) : Return pull : End Function ; OKAY

Function bfs(flag) ; 99%
If flagset(flag) Then pc=pc+signedbyte(peek(pc))+1 Else pc=pc+1
pc=Abs (pc+65536) And 65535
End Function

Function bfc(flag) ; 99%
If flagset(flag) Then pc=pc+1 Else pc=pc+signedbyte(peek(pc))+1
pc=Abs (pc+65536) And 65535
End Function

Function adc(address) ; 99% ; DECIMAL MODE NOT NEEDED?
val=peek(address) : h=a+val : If flagset(c) Then h=h+1
a=h : setflag(c,h>$FF) : a=Abs(a+256) And 255 : setflag(z,a=0) : setflag(n,a And n)
h=h+$80
setflag(v, ( (h>$FF) Or (h<0)) ) ; SetFlag(v,( (((a Xor val) And $80) And ((a Xor h) And  $80)) Xor 65535) And 65536 )
End Function

Function sbc (address) ; 99% ; DECIMAL MODE NOT NEEDED?
val=peek(address) : h=a-val : If Not flagset(c) Then h=h-1
a=h :setflag(c,h<=$FF) : a=Abs(a+256) And 255 : setflag(z,a=0) : setflag(n,a And n)
h=h+$80
setflag(v, ((h>$FF) Or (h<0)) ) ; SetFlag(v,(((a Xor val) And $80) And ((a Xor h) And $80)) And 65536)
End Function

Function asl(address) ; 99%
byte=peek(address) : setflag(c,(byte And $80)) : byte=byte Shl 1 : byte=byte And 255 : poke(address,byte) : setflag(z,byte=0) : setflag(n,byte And n)
End Function

Function rol_A() ; 99%
bit=flagset(c) : setflag(c,(a And $80)) : a=a Shl 1 : a=a And 255
If bit Then a=a Or $01
setflag(z,a=0) : setflag(n,a And n)
End Function

Function rol(address) ; 99%
byte=peek(address) : bit=flagset(c) : setflag(c,(byte And $80)) : byte=byte Shl 1 : byte=byte And 255
If bit Then byte=byte Or $01
poke(address,byte) : setflag(z,byte=0) : setflag(n,byte And n)
End Function

Function ror_A() ; 99%
bit=flagset(c) : setflag(c,(a And $01)) : a=a Shr 1 : a=a And 255
If bit Then a=a Or $80
setflag(z,a=0) : setflag(n,a And n)
End Function

Function ror(address) ; 99%
byte=peek(address) : bit=flagset(c) : setflag(c,(byte And $01)) : byte=byte Shr 1 : byte=byte And 255
If bit Then byte=byte Or $80
poke(address,byte) : setflag(z,byte=0) : setflag(n,byte And n)
End Function

Function lsr_A() : setflag(c,(a And $01)) : a=a Shr 1 : a=a And 255 : setflag(z,a=0) : setflag(n,False) : End Function ; OKAY
Function lsr(address) : byte=peek(address) : setflag(C,(byte And $01)) : byte=byte Shr 1 : byte=byte And 255 : poke(address,byte) : setflag(z,byte=0) : setflag(n,False) : End Function ; 99%
Function bit(address) : h=peek(address) : setflag(n,(h And $80)) : setflag(v,(h And $40)) : setflag(z,((h And a)=0) ) : End Function ; 99%
Function brk() : setflag(b,True) : pc=pc+1 : pc=pc And 65535 : push(hi(pc)) : push(lo(pc)) : push(p) : setflag(i,True) : pc=peek2($FFFE) : End Function ; 99%
Function and_(address) : a=a And peek(address) : setflag(z,a=0) : setflag(n,a And n) : End Function ; OKAY!
Function ora(address) : a=a Or peek(address) : setflag(z,a=0) : setflag(n,a And n) : End Function ; OKAY!
Function asl_A() : setflag(c,(a And $80)) : a=a Shl 1 : a=a And 255 : setflag(z,a=0) : setflag(n,a And n) : End Function ; 99%
Function cmp(address) : h=a-peek(address) : setflag(c,h<=$FF) : h=Abs(h+256) And 255 : setflag(z,h=0) : setflag(n,h And n) : End Function ; 99% + 256 as could be negative? carry measuring correct?
Function cpx(address) : h=x-peek(address) : setflag(c,h<=$FF) : h=Abs(h+256) And 255 : setflag(z,h=0) : setflag(n,h And n) : End Function ; 99% + 256 as could be negative? carry measuring correct?
Function cpy(address) : h=y-peek(address) : setflag(c,h<=$FF) : h=Abs(h+256) And 255 : setflag(z,h=0) : setflag(n,h And n) : End Function ; 99% + 256 as could be negative? carry measuring correct?
Function dec_(address) :h=peek(address)-1 : h=Abs(h+256) And 255 : poke (address,h) : setflag(z,h=0) : setflag(n,h And n) : End Function ; 99%
Function dex() : x=x-1 : x=Abs(x+256) And 255 : setflag(z,x=0) : setflag(n,x And n) : End Function ; 99%
Function dey() : y=y-1 : y=Abs(y+256) And 255 : setflag(z,y=0) : setflag(n,y And n) : End Function ; 99%
Function eor(address) : a=a Xor peek(address) : setflag(z,a=0) : setflag(n,a And n) : End Function ; OKAY!
Function inc_(address) : h=peek(address)+1 : h=h And 255 : poke (address,h) : setflag(z,h=0) : setflag(n,h And n) : End Function ; 99%
Function inx() : x=(x+1) And 255 : setflag(z,x=0) : setflag(n,x And n) : End Function ; OKAY
Function iny() : y=(y+1) And 255 : setflag(z,y=0) : setflag(n,y And n) : End Function ; OKAY
Function jmp(address) : pc=address : End Function ; OKAY
Function jsr(address) : pc=pc-1 : pc=Abs(pc+65536) And 65535 : push(hi(pc)) : push(lo(pc)) : pc=address : End Function ; OKAY
Function rts() : pc=pull() : pc=pc+pull()*256+1 : End Function ; OKAY!
Function rti() : p=pull() : pc=pull() : pc=pc+pull()*256 : End Function ; OKAY!
Function lda(address) : a=peek(address) : setflag(z,a=0) : setflag(n,a And n) : End Function ; OKAY!
Function ldx(address) : x=peek(address) : setflag(z,x=0) : setflag(n,x And n) : End Function ; OKAY!
Function ldy(address) : y=peek(address) : setflag(z,y=0) : setflag(n,y And n) : End Function ; OKAY!
Function pla() : a=pull() : setflag(z,a=0) : setflag(n,a And n) : End Function ; OKAY
Function sta(address) : poke (address,a) : End Function ; OKAY!
Function stx(address) : poke (address,x) : End Function ; OKAY!
Function sty(address) : poke (address,y) : End Function ; OKAY!
Function tax() : x=a : setflag(z,x=0) : setflag(n,x And n) : End Function ; OKAY!
Function tay() : y=a : setflag(z,y=0) : setflag(n,y And n) : End Function ; OKAY!
Function tsx() : x=s : setflag(z,x=0) : setflag(n,x And n) : End Function ; OKAY!
Function txa() : a=x : setflag(z,a=0) : setflag(n,a And n) : End Function ; OKAY!
Function txs() : s=x : End Function ; OKAY!
Function tya() : a=y : setflag(z,a=0) : setflag(n,a And n) : End Function ; OKAY!

Function vicPeek(address) ; returns byte OKAY
peek=ram(address)
If address=>$1000 And address=<$1FFF Then peek=char_rom(address+$C000)
If address=>$9000 And address=<$9FFF Then peek=char_rom(address+$4000)
Return peek
End Function

Function vicText(pos,val) ; OKAY
LockBuffer() : x0=(pos Mod 40)*8 : y0=Int((pos / 40))*8
For xxx=0 To 7 : For yyy=0 To 7
colour=$352879 : If (vicPeek($1000+val*8+yyy) And ($80 Shr xxx)) Then colour = $6C5EB5
WritePixelFast (x0+xxx) Shl 1,(y0+yyy) Shl 1,colour : WritePixelFast 1+(x0+xxx) Shl 1,(y0+yyy) Shl 1,colour ; with abit retro feeling
Next : Next : UnlockBuffer()
End Function

Function reset() : pc=peek2($FFFC) : setflag(i,True) : End Function ; OKAY

Function irq() ; VERTCIAL BLANKING INTERUPT ; OKAY
If Not flagset(i) Then setflag(B,False) : push(hi(pc)) : push(lo(pc)) : push(p) : setflag(i,True) : pc=peek2($FFFE) : addkey()
End Function

